import os
from shutil import rmtree

import numpy as np
import pytest
from ase.build import bulk, molecule
from ase.calculators.emt import EMT
from ase.calculators.lj import LennardJones
from ase.optimize import BFGS, BFGSLineSearch

from quacc import SETTINGS
from quacc.runners.ase import run_calc, run_opt, run_vib

DEFAULT_SETTINGS = SETTINGS.model_copy()


def prep_files():
    # Make some test files to play with
    if not os.path.exists("test_calc"):
        os.mkdir("test_calc")
    with open("test_file.txt", "w") as f:
        f.write("test")


def teardown_function():
    if os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_calc")):
        rmtree(os.path.join(SETTINGS.RESULTS_DIR, "test_calc"), ignore_errors=True)
    for f in ["test_file.txt", "test_file.txt.gz"]:
        if os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, f)):
            os.remove(os.path.join(SETTINGS.RESULTS_DIR, f))


def test_run_calc(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    prep_files()

    atoms = bulk("Cu") * (2, 1, 1)
    atoms[0].position += 0.1
    atoms.calc = EMT()

    new_atoms = run_calc(atoms, copy_files=["test_file.txt"])
    assert atoms.calc.results is not None
    assert new_atoms.calc.results is not None
    assert not os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt"))
    assert os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt.gz"))
    assert np.array_equal(new_atoms.get_positions(), atoms.get_positions()) is True
    assert np.array_equal(new_atoms.cell.array, atoms.cell.array) is True


def test_run_calc_no_gzip(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    prep_files()

    SETTINGS.GZIP_FILES = False

    atoms = bulk("Cu") * (2, 1, 1)
    atoms[0].position += 0.1
    atoms.calc = EMT()

    new_atoms = run_calc(atoms, copy_files=["test_file.txt"])
    assert atoms.calc.results is not None
    assert new_atoms.calc.results is not None
    assert os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt"))
    assert not os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt.gz"))
    assert np.array_equal(new_atoms.get_positions(), atoms.get_positions()) is True
    assert np.array_equal(new_atoms.cell.array, atoms.cell.array) is True
    SETTINGS.GZIP_FILES = DEFAULT_SETTINGS.GZIP_FILES


def test_run_opt1(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    prep_files()

    atoms = bulk("Cu") * (2, 1, 1)
    atoms[0].position += 0.1
    atoms.calc = EMT()

    dyn = run_opt(atoms, copy_files=["test_file.txt"])
    traj = dyn.traj_atoms
    assert traj[-1].calc.results is not None
    assert not os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt"))
    assert os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt.gz"))
    assert np.array_equal(traj[-1].get_positions(), atoms.get_positions()) is False
    assert np.array_equal(traj[-1].cell.array, atoms.cell.array) is True


def test_run_opt2(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    atoms = bulk("Cu") * (2, 1, 1)
    atoms[0].position += 0.1
    atoms.calc = EMT()

    dyn = run_opt(
        atoms,
        optimizer=BFGS,
        copy_files=["test_file.txt"],
        optimizer_kwargs={"restart": None},
    )
    traj = dyn.traj_atoms
    assert traj[-1].calc.results is not None

    dyn = run_opt(
        traj[-1],
        optimizer=BFGSLineSearch,
        copy_files=["test_file.txt"],
        optimizer_kwargs={"restart": None},
    )
    traj = dyn.traj_atoms
    assert traj[-1].calc.results is not None


def test_run_vib(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    prep_files()

    o2 = molecule("O2")
    o2.calc = LennardJones()
    vib = run_vib(o2, copy_files=["test_file.txt"])
    assert np.real(vib.get_frequencies()[-1]) == pytest.approx(255.6863883406967)
    assert np.array_equal(vib.atoms.get_positions(), o2.get_positions()) is True
    assert not os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt"))
    assert os.path.exists(os.path.join(SETTINGS.RESULTS_DIR, "test_file.txt.gz"))


def test_bad_runs(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    atoms = bulk("Cu")
    atoms.calc = EMT()

    # No file
    with pytest.warns(UserWarning):
        run_calc(atoms, copy_files=["test_file.txt"])

    # No file again
    with pytest.warns(UserWarning):
        run_opt(atoms, copy_files=["test_file.txt"])

    # No trajectory kwarg
    with pytest.raises(ValueError):
        run_opt(
            atoms,
            optimizer=BFGSLineSearch,
            optimizer_kwargs={"restart": None, "trajectory": "test.traj"},
        )


def test_unique_workdir(tmp_path, monkeypatch):
    SETTINGS.CREATE_UNIQUE_WORKDIR = True
    monkeypatch.chdir(tmp_path)
    prep_files()

    # Static
    atoms = bulk("Cu") * (2, 1, 1)
    atoms[0].position += 0.1
    atoms.calc = EMT()

    run_calc(atoms, copy_files=["test_file.txt"])
    assert atoms.calc.results is not None

    # Opt
    atoms = bulk("Cu") * (2, 1, 1)
    atoms[0].position += 0.1
    atoms.calc = EMT()

    dyn = run_opt(atoms, copy_files=["test_file.txt"])
    traj = dyn.traj_atoms
    assert traj[-1].calc.results is not None
    assert np.array_equal(traj[-1].get_positions(), atoms.get_positions()) is False
    assert np.array_equal(traj[-1].cell.array, atoms.cell.array) is True

    # Vib
    o2 = molecule("O2")
    o2.calc = LennardJones()
    vib = run_vib(o2)
    assert np.real(vib.get_frequencies()[-1]) == pytest.approx(255.6863883406967)
    assert np.array_equal(vib.atoms.get_positions(), o2.get_positions()) is True

    SETTINGS.CREATE_UNIQUE_WORKDIR = DEFAULT_SETTINGS.CREATE_UNIQUE_WORKDIR
